#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .settings import Settings
from .utils.parse_excel import ExcelParser
from .utils.logger import logger
from importlib import import_module
import unittest
import re


class TestSet:
    """读取测试用例表格中的数据，生成PUnittest可识别的测试用例对象"""
    def __init__(self, filter_hidden=True):
        """
        :param filter_hidden: 过滤被隐藏的excel
        """
        self._excel = ExcelParser(Settings.TEST_EXCEL_PATH, 'TestCases')
        self.tags = Settings.RUN_TAGS
        self.excel_cases = self._get_excel_cases(True)
        self.test_suite = None

    @staticmethod
    def _convert_params(params):
        """将测试用例中的ParamsData由字符串转换成python数据结构"""
        if params is None:
            return None
        elif isinstance(params, bool):
            return params
        else:
            param_list = params.split('\n')
            try:
                param_list = [eval(x.strip()) for x in param_list]
            except NameError:
                param_list = [x.strip() for x in param_list]
            return param_list

    @staticmethod
    def _convert_asserts(asserts):
        """将测试用例中的AssertResult由字符串转换成python数据结构"""
        if asserts is None:
            return None
        elif isinstance(asserts, bool):
            return asserts
        else:
            asserts_list = asserts.split('\n')
            try:
                asserts_list = [eval(x.strip()) for x in asserts_list]
            except NameError:
                asserts_list = [x.strip() for x in asserts_list]
            return asserts_list

    @staticmethod
    def _convert_tags(tags):
        """将测试用例中的Tag解析成列表"""
        if tags is None:
            return ['All']
        else:
            tag_list = ['All']
            tag_list.extend(tags.split('\n'))
            param_list = [x.strip() for x in tag_list]
            return param_list

    def _get_excel_cases(self, filter_hidden):
        """
        一次性读取所有测试用例
        :param filter_hidden: bool, 是否过滤excel中被隐藏的行
        :return: [dict(case), dict(case)...]形式的suite
        """
        test_set = list()
        titles = [x for x in self._excel.get_row_values(1)]
        for row in range(2, self._excel.max_row + 1):
            test_case = dict()
            # 获取该行是否被隐藏
            row_hidden = self._excel.worksheet.row_dimensions[row].hidden
            if filter_hidden is False or (filter_hidden is True and row_hidden is False):
                values = [x for x in self._excel.get_row_values(row)]
                if len(values) < len(titles):
                    for i in range(len(titles) - len(values)):
                        values.append(None)
                for idx, (title, value) in enumerate(zip(titles, values)):
                    if title == 'ParamsData':
                        params = self._convert_params(value)
                        test_case[title] = params
                    elif title == 'AssertResult':
                        asserts = self._convert_asserts(value)
                        test_case[title] = asserts
                    elif title == 'Tags':
                        tags = self._convert_tags(value)
                        test_case[title] = tags
                    else:
                        test_case[title] = value
                test_set.append(test_case)
        self._test_set = test_set
        return test_set

    def filter_test_case(self, **titles):
        """
        根据表格中的title过滤测试用例
        :param titles: 关键字参数，每个参数名对应表格中的title
        :return: 测试用例组成的列表
        """
        test_case = self.test_set
        if titles:
            for title, value in titles.items():
                test_case = list(filter(lambda d: title in d.keys(), test_case))
                test_case = list(filter(lambda d: d[title] == value, test_case))
        return test_case

    def make_test_suite(self):
        """
        将excel表格中读取的测试用例生成 PUnittest 能识别的 TestSuite 对象，或者获取本地的所有测试文件
        :return: unittest.TestSuite 对象
        """
        test_suite = unittest.TestSuite()
        # 从Excel中读取测试用例
        if Settings.EXCEL_TEST_SET:
            if len(self.excel_cases) > 0:
                for test_case in self.excel_cases:
                    _dir_name = test_case['TestDir'] if 'TestDir' in test_case else None
                    _file_name = test_case['TestFile'] if 'TestFile' in test_case else None
                    _cls_name = test_case['TestClass'] if 'TestClass' in test_case else None
                    _case_name = test_case['TestCase'] if 'TestCase' in test_case else None
                    try:
                        package = import_module('{0}.{1}'.format(_dir_name, _file_name))
                        cls = getattr(package, _cls_name)
                        for name, func in list(cls.__dict__.items()):
                            _name = re.sub(r'_#[\d]*', '', name)
                            _name = re.sub(r'test_\d{5}', 'test', _name)
                            if _name == _case_name:
                                case = cls(name)
                                test_suite.addTest(case)
                    except Exception as e:
                        logger.error('Fail to load test case <{0}><{1}>: {2}'.format(_cls_name, _case_name, e))
            else:
                logger.error('Fail to load any test case, please check')
                raise RuntimeError('Fail to load any test case, please check')
        # 从TestSuite文件夹中读取所有测试用例
        else:
            test_suite.addTests(unittest.TestLoader().discover(Settings.TEST_SUITE_DIR))
        self.test_suite = test_suite
        return test_suite


if __name__ == '__main__':
    t = TestSet()
    suite = t.make_test_suite()
    runner = unittest.TextTestRunner()
    runner.run(suite)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .logger import logger
from openpyxl.styles import PatternFill
import openpyxl


def create_excel_file(file_path):
    """
    创建xlsx表格
    :param file_path: 表格全路径
    :return: None
    """
    wb = openpyxl.Workbook()
    wb.create_sheet(title="Sheet1", index=0)
    wb.save(file_path)


class ExcelParser:

    def __init__(self, excel_file_path, sheet_name):
        self.excel_file_path = excel_file_path
        self.workbook = openpyxl.load_workbook(self.excel_file_path)
        self.worksheet = self.workbook[sheet_name]
        self.max_row = self.worksheet.max_row
        self.max_column = self.worksheet.max_column

    def save_workbook(self):
        self.workbook.save(self.excel_file_path)

    def create_new_sheet(self, new_sheet_name="new_sheet"):
        try:
            self.workbook.create_sheet(new_sheet_name)
            self.workbook.save(self.excel_file_path)
        except Exception as e:
            logger.error("Fail to create new sheet: {0}".format(e)[:-1])
            raise e

    def get_cell_value(self, row, column):
        return self.worksheet.cell(row=row, column=column).value

    def set_cell_value(self, row, column, value):
        self.worksheet.cell(row=row, column=column).value = value
        return value

    def get_row_values(self, row, filter_none='end'):
        row_values = [self.get_cell_value(row, col) for col in range(1, self.max_column + 1)]
        if filter_none in ['all', 'All', 'ALL']:
            row_values = list(filter(lambda value: value is not None, row_values))
        elif filter_none in ['end', 'End', 'END']:
            for i, val in enumerate(row_values[::-1]):
                if val is not None:
                    row_values = row_values[:len(row_values)-i]
                    break
        return row_values

    def get_col_values(self, col, filter_none='end'):
        col_values = [self.get_cell_value(row, col) for row in range(1, self.max_row + 1)]
        if filter_none in ['all', 'All', 'ALL']:
            col_values = list(filter(lambda value: value is not None, col_values))
        elif filter_none in ['end', 'End', 'END']:
            for i, val in enumerate(col_values[::-1]):
                if val is not None:
                    col_values = col_values[:len(col_values)-i]
                    break
        return col_values

    def set_cell_background_color(self, row, column, fgColor):
        fill = PatternFill(fgColor=fgColor, fill_type='solid')
        self.worksheet.cell(row=row, column=column).fill = fill
        return True


if __name__ == '__main__':
    pass



#! /usr/bin/env python
# -*- coding: UTF-8 -*-

from .settings import Settings
from .testset import TestSet
from .utils.logger import logger
from functools import wraps
import unittest
import sys

CASE_TAG_FLAG = "__case_tag__"
CASE_DATA_FLAG = "__case_data__"
CASE_ID_FLAG = "__case_id__"
CASE_INFO_FLAG = "__case_info__"
CASE_RUN_INDEX_FlAG = "__case_run_index_flag__"
CASE_SKIP_FLAG = "__unittest_skip__"
CASE_SKIP_REASON_FLAG = "__unittest_skip_why__"


def skip(reason):
    def wrap(func):
        return unittest.skip(reason)(func)
    return wrap


def skip_if(condition, reason):
    def wrap(func):
        return unittest.skipIf(condition, reason)(func)
    return wrap


def data(params, asserts):
    """注入测试数据，可以做为测试用例的数据驱动
    :param params: 测试数据组成的list
    :param asserts: 断言数据组成的list
    :return:
    """
    def wrap(func):
        param_data = {'params': params, 'asserts': asserts}
        setattr(func, CASE_DATA_FLAG, param_data)
        return func
    return wrap


def tag(*tag_type):
    """指定测试用例的标签，可以作为测试用例分组使用，用例默认会有Tag.ALL标签，支持同时设定多个标签，如：
    @tag(Tag.V1_0_0, Tag.SMOKE)
    def test_func(self):
        pass

    :param tag_type:标签类型，在tag.py里边自定义
    :return:
    """
    def wrap(func):
        if not hasattr(func, CASE_TAG_FLAG):
            tags = {'All'}
            tags.update(tag_type)
            setattr(func, CASE_TAG_FLAG, tags)
        else:
            getattr(func, CASE_TAG_FLAG).update(tag_type)
        return func
    return wrap


def _assertion(func):
    """用来装饰PUnittest类中所有的AssertXxx方法，提供日志，
    并且当出现断言异常不会立即终止用例，而是等用例结束后抛出所有的断言异常"""
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            logger.info('[Assert]: {0}{1}'.format(func.__name__, format(args[1:])))
            return func(*args, **kwargs)
        except AssertionError as e:
            args[0].Exc_Stack.append(e)
    return wrapper


def _handler(rerun=Settings.CASE_FAIL_RERUN):
    """处理测试用例，为用例添加格式化的日志。并提供失败重跑机制。"""
    def wrapper(func):
        @wraps(func)
        def wrap(*args, **kwargs):
            _testcase_name = args[0]._testMethodName
            _testcase_doc = args[0]._testMethodDoc
            _testclass_name = args[0].__class__.__name__
            _testclass_doc = args[0].__class__.__doc__
            rerun_time = rerun
            while rerun_time > 0:
                try:
                    logger.info('[TestCase]: {0}.{1} {2}|{3}'.format(
                        _testclass_name, _testcase_name, _testclass_doc, _testcase_doc))
                    logger.info('[TestProgress]: {0}/{1} [RerunTime]: NO.{2}'.format(
                        getattr(func, CASE_RUN_INDEX_FlAG), TestCaseHandler.actual_case_num, rerun - rerun_time + 1))
                    args[0]._clear_exc_stack()
                    result = func(*args, **kwargs)
                    # 用例执行完毕抛出所有可能存在的AssertionError异常
                    args[0]._raise_exc()
                    logger.info('[TestResult]: Pass!')
                    return result
                except AssertionError as e:
                    rerun_time -= 1
                    if rerun_time <= 0:
                        # exc_type, exc_msg, exc_traceback = sys.exc_info()
                        logger.error('[TestResult]: Fail!')
                        # logger.error('[FailReason]: {0}: {1}'.format(exc_type.__name__, exc_msg))
                        raise e
                except Exception as e:
                    rerun_time -= 1
                    if rerun_time <= 0:
                        _, _, exc_traceback = sys.exc_info()
                        logger.error('[TestResult]: Error!')
                        # import traceback
                        # for t in traceback.format_exc().splitlines():
                        #     logger.error('[Traceback]: {0}'.format(t))
                        raise e
        return wrap
    return wrapper


def _feed_data(*args, **kwargs):
    def wrap(func):
        @wraps(func)
        def _wrap(self):
            return func(self, *args, **kwargs)
        return _wrap
    return wrap


class TestCaseHandler:
    actual_case_num = 0
    total_case_num = 0

    @classmethod
    def create_case_id(cls):
        cls.total_case_num += 1
        return cls.total_case_num

    @classmethod
    def create_actual_run_index(cls):
        cls.actual_case_num += 1
        return cls.actual_case_num

    @staticmethod
    def modify_func_name(func, index):
        """修改函数名字，实现排序，并且根据DDT数据的编号在函数名后添加编号 eg test_fight >> test_00001_fight_#1
        :param func:
        :param index:
        :return:
        """
        case_id = TestCaseHandler.create_case_id()
        setattr(func, CASE_ID_FLAG, case_id)
        func_name = func.__name__.replace("test_", "test_{:05d}_".format(case_id))
        if index is not None:
            func_name += "_#{0}".format(index)
        return func_name

    @staticmethod
    def create_case_with_case_data(func):
        """根据原有测试函数创建新的带参数的测试函数"""
        result = dict()
        test_data = getattr(func, CASE_DATA_FLAG)
        params = test_data['params']
        asserts = test_data['asserts']
        for index, (param, assert_) in enumerate(zip(params, asserts), 1):
            if not hasattr(func, CASE_SKIP_FLAG):
                setattr(func, CASE_RUN_INDEX_FlAG, TestCaseHandler.create_actual_run_index())
            func_name = TestCaseHandler.modify_func_name(func, index)
            result[func_name] = _handler()(_feed_data(params=param, asserts=assert_)(func))
        return result

    @staticmethod
    def create_case_without_case_data(func):
        """根据原有测试函数创建新的不带参数的测试函数"""
        result = dict()
        if not hasattr(func, CASE_SKIP_FLAG):
            setattr(func, CASE_RUN_INDEX_FlAG, TestCaseHandler.create_actual_run_index())
        func_name = TestCaseHandler.modify_func_name(func, None)
        result[func_name] = _handler()(func)
        return result

    @staticmethod
    def classify_funcs(funcs_dict):
        """将类中的函数进行归类"""
        funcs = dict()
        cases = dict()
        asserts = dict()
        for attr in funcs_dict:
            if attr.startswith("test_"):
                cases[attr] = funcs_dict[attr]
            elif attr.startswith("assert"):
                asserts[attr] = funcs_dict[attr]
            else:
                funcs[attr] = funcs_dict[attr]
        return funcs, cases, asserts

    @staticmethod
    def filter_cases_by_tags(tags, func):
        """根据func的tags进行过滤"""
        name = func.__name__
        if name.startswith('test_'):
            _tags = getattr(func, CASE_TAG_FLAG, 'All')
            for _tag in _tags:
                if _tag in tags:
                    break
            else:
                msg = "不包含需要执行的Tag: {}".format(', '.join(tags))
                filtered_func = skip(msg)(func)
                return filtered_func
        return func


class Meta(type):
    def __new__(mcs, cls_name, supers, cls_dict):
        testset = TestSet()
        sets = testset.excel_cases
        tags = testset.tags
        funcs, cases, asserts = TestCaseHandler.classify_funcs(cls_dict)
        for name, assert_ in asserts.items():
            # 将装饰过的assert方法加入类
            funcs.update({name: _assertion(assert_)})

        # 读取Excel中的数据为测试函数绑定装饰器
        if Settings.EXCEL_TEST_SET:
            for test_case in sets:
                _cls_name = test_case['TestClass'] if 'TestClass' in test_case else None
                _case_name = test_case['TestCase'] if 'TestCase' in test_case else None
                _case_desc = test_case['CaseDescription'] if 'CaseDescription' in test_case else None
                _tags = test_case['Tags'] if 'Tags' in test_case else None
                _skip = test_case['Skip'] if 'Skip' in test_case else None
                _params = test_case['ParamsData'] if 'ParamsData' in test_case else None
                _asserts = test_case['AssertResult'] if 'AssertResult' in test_case else None
                if cls_name == _cls_name:
                    for test_func in cases.values():
                        if test_func.__name__ == _case_name:
                            # 为测试函数绑定tag装饰器
                            test_func = tag(*_tags)(test_func)
                            # 根据tag进行过滤
                            test_func = TestCaseHandler.filter_cases_by_tags(tags, test_func)
                            # 为测试函数绑定skip装饰器，excel和硬编码任何一个有skip则跳过
                            skip_ = getattr(test_func, CASE_SKIP_FLAG, None)
                            if not skip_ and _skip:
                                test_func = skip(_skip)(test_func)
                            # 修改测试函数的文档注释
                            if _case_desc:
                                setattr(test_func, '__doc__', _case_desc)
                            # 为测试函数绑定data装饰器
                            if _asserts is not None and _params is not None:
                                test_func = data(_params, _asserts)(test_func)
                                funcs.update(TestCaseHandler.create_case_with_case_data(test_func))
                            else:
                                funcs.update(TestCaseHandler.create_case_without_case_data(test_func))

        # 根据测试函数硬编码为测试函数绑定装饰器
        else:
            for test_func in cases.values():
                if not hasattr(test_func, CASE_TAG_FLAG):
                    setattr(test_func, CASE_TAG_FLAG, 'All')  # 没有指定tag的用例，默认带有tag：ALL
                # 根据tag进行过滤
                test_func = TestCaseHandler.filter_cases_by_tags(tags, test_func)
                # 注入用例信息
                case_info = "{}.{}".format(test_func.__module__, test_func.__name__)
                setattr(test_func, CASE_INFO_FLAG, case_info)
                # 注入测试数据
                if hasattr(test_func, CASE_DATA_FLAG):
                    funcs.update(TestCaseHandler.create_case_with_case_data(test_func))
                else:
                    funcs.update(TestCaseHandler.create_case_without_case_data(test_func))

        return super(Meta, mcs).__new__(mcs, cls_name, supers, funcs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from punittest import PUnittest, logger


class TestChat(PUnittest):
    """测试聊天功能"""

    def setUp(self):
        super(TestChat, self).setUp()
        logger.info("登录角色")

    def tearDown(self):
        logger.info("角色登出")
        super(TestChat, self).tearDown()

    @PUnittest.tag("Smoke", "Regression")
    @PUnittest.data(["哈哈", "此处超过字数上限"], [True, False])
    def test_chat(self, params, asserts):
        """测试在线聊天"""
        logger.info("登录另一个账号")
        logger.info("调用聊天接口，发送聊天信息 {}".format(params))
        result = True
        self.assertEqual(result, asserts)

    @PUnittest.tag("Smoke", "Regression")
    @PUnittest.data(["哈哈"], ["聊天对象不在线"])
    def test_msg(self, params, asserts):
        """测试留言"""
        logger.info("调用留言接口，发送留言信息 {}".format(params))
        result = "聊天对象不在线"
        self.assertIn(result, asserts)
